#ifndef LATTICEPROTEINPIVOTMOVESTRUCTSTATE_HH_
#define LATTICEPROTEINPIVOTMOVESTRUCTSTATE_HH_


#include "LatticeProteinPivotMoveState.hh"

namespace ell
{
		/**
		 * This class generates neighbors of lattice proteins in the
		 * energy landscape by applying pivot moves.
		 * All lattice proteins of this class have a common sequence,
		 * contact energy function and lattice model.
		 *
		 * @author Martin Mann, Sebastian Will, Andreas Richter
		 */
	class LatticeProteinPivotMoveStructState : public LatticeProteinPivotMoveState
	{
	protected:	
		class NeighborList : public LatticeProteinPivotMoveState::NeighborList
		{
		protected:
				//! the state of origin of the neighbors
			const LatticeProteinPivotMoveStructState* const origin;

		public:
			NeighborList(const LatticeProteinPivotMoveStructState* _origin);

				/*! Returns a pointer to the first element of the
				 *  virtual list of neighbors of the state of origin.
				 *  Returns NULL if no neighbors exist.
				 */
			virtual State* first(State::NeighborList::ItState** itstate) const;
						 
				/*! Returns a pointer to the next element of the
				 *  virtual list of neighbors of the state of origin.
				 *  Returns NULL if no more neighbors exist.
				 */
			virtual State* next(	State::NeighborList::ItState* itstate, 
									State* elem) const;
		};

		class RandomNeighborList : public LatticeProteinPivotMoveState::
											RandomNeighborList
		{
		protected:
				//! the state of origin of the neighbors
			const LatticeProteinPivotMoveStructState* const origin;

		public:
			RandomNeighborList(const LatticeProteinPivotMoveStructState* _origin);

				/*! Returns a pointer to a random element of the
				 *  virtual list of neighbors of the state of origin.
				 *  Returns NULL if no neighbors exist.
				 */
			virtual State* first(State::NeighborList::ItState** itstate) const;
						 
				/*! Returns a pointer to the next random element of the
				 *  virtual list of neighbors of the state of origin.
				 *  Returns NULL if no more neighbors exist.
				 */
			virtual State* next(	State::NeighborList::ItState* itstate, 
									State* elem) const;
		};

			/*! common lattice model */
		static const LatticeModel*	commonLattice;
		
			/*! common contact energy function */
		static const ContactEnergyFunction*	commonEnergy;	

			//! common sequence string
		static std::string				commonSeqStr;

	public:
			 /*! The commonsequence, contact energy function and
			  *  lattice model have to be set before calling the constructor.
			  */
		LatticeProteinPivotMoveStructState(const std::string& moveString, 
										   const bool isAbsoluteMove);
				
		LatticeProteinPivotMoveStructState( 
						const LatticeProteinPivotMoveStructState& latProtPMState2);
						
		virtual ~LatticeProteinPivotMoveStructState();

			/*! Sets common lattice model.
			 *
			 *  Has to be called before initialisation of an
			 *  LatticeProteinPivotMoveStructState object and
			 *  has to be called only once.
			 */
		static void		setCommonLatticeModel(const LatticeModel* lattice);

			/*! Sets common contact energy function.
			 *
			 *  Has to be called before initialisation of an
			 *  LatticeProteinPivotMoveStructState object and
			 *  has to be called only once.
			 */
		static void		setCommonContactEnergyFunction(
						 const ContactEnergyFunction* energy);

			/*! Sets common sequence string.
			 *
			 *  Has to be called before initialisation of an
			 *  LatticeProteinPivotMoveStructState object and
			 *  has to be called only once.
			 */
		static void		setCommonSequenceString(const std::string& seqStr);

			/*! Returns the common lattice model.
			 */
		static const LatticeModel*			getCommonLatticeModel() {
			return commonLattice;
		}

			/*! Returns the contact energy function.
			 */
		static const ContactEnergyFunction*	getCommonContactEnergyFunction() {
			return commonEnergy;
		}

			/*! Returns the common sequence string.
			 */
		static std::string					getCommonSequenceString() {
			return commonSeqStr;
		}

	// abstract functions (State)
			/*! Returns a pointer to a clone of the current state. */
		virtual LatticeProteinPivotMoveStructState*	clone() const;
		
		virtual CState* compressedClone() const;

		virtual void	compress(CState& cState) const;

		virtual NeighborListPtr getNeighborList() const;
		
		virtual NeighborListPtr getRandomNeighborList() const;

	};

} // namespace ell

#endif /*LATTICEPROTEINPIVOTMOVESTRUCTSTATE_HH_*/
