#include <ell/LatticeProtein.hh>

namespace ell
{

	LatticeProtein::LatticeProtein( const LatticeModel* _lattice, 
		const ContactEnergyFunction* _energy, const std::string& seqStr, 
		const std::string& moveString, const bool isAbsoluteMove)
	 :	lattice(_lattice), energy(_energy), sequence(NULL), relMoves(NULL)
	{
		assertell (lattice != NULL && energy != NULL, 
			"no lattice model or energy function available");
			// sequence initialisieren
		sequence = new Sequence(energy->getAlphabet()->getSequence(seqStr));
			// relative move sequence initialisieren 
			// je nachdem ob absoluter oder relativer string uebergeben
		if (isAbsoluteMove) {
			relMoves = new MoveSequence(
							lattice->absMovesToRelMoves(
								lattice->parseMoveString(moveString)));
		} else {
			relMoves = new MoveSequence(lattice->parseMoveString(moveString));
		}
		assertell ((relMoves->size()+1) == sequence->size(), 
			"sequence and structure length differ"); // muessen gleich lang sein
	}

	LatticeProtein::LatticeProtein(const LatticeProtein& latPr) 
	 :	lattice(latPr.lattice), energy(latPr.energy), 
	 	sequence(new Alphabet::Sequence(*latPr.sequence)), 
	 	relMoves(new MoveSequence(*(latPr.relMoves)))
	{
		assertell (lattice != NULL && energy != NULL, 
			"no lattice model or energy function available");
		assertell ((relMoves->size()+1) == sequence->size(), 
			"sequence and structure length differ"); // muessen gleich lang sein
	}
	
	LatticeProtein::~LatticeProtein()
	{
		if (sequence != NULL) {
			delete sequence;	sequence = NULL;
		}
		if (relMoves != NULL) {
			delete relMoves;	relMoves = NULL;
		}
	}
	
	
	LatticeProtein&
	LatticeProtein::operator =(const LatticeProtein& latProt2) {
		if (this != &latProt2) {
			lattice = latProt2.lattice;
			energy = latProt2.energy;
			if (sequence != NULL) {
				delete sequence;	sequence = NULL;
			}
			sequence = new Sequence(*(latProt2.sequence));
			if (relMoves != NULL) {
				delete relMoves;	relMoves = NULL;
			}
			relMoves = new MoveSequence(*(latProt2.relMoves));
		}
		return *this;
	}
	
	bool
	LatticeProtein::operator ==(const LatticeProtein& latProt2) const {
		return	*lattice == *(latProt2.lattice)
				&& *energy == *(latProt2.energy)
				&& *sequence == *(latProt2.sequence)
				&& *relMoves == *(latProt2.relMoves);
	}
	
	bool
	LatticeProtein::operator !=(const LatticeProtein& latProt2) const {
		return	*lattice != *(latProt2.lattice)
				|| *energy != *(latProt2.energy)
				|| *sequence != *(latProt2.sequence)
				|| *relMoves != *(latProt2.relMoves);
	}
	
	
	bool 
	LatticeProtein::isSelfAvoiding() const {
		assertell(relMoves != NULL, "no structure available");
		bool retVal = true;
		IPointVec pData = lattice->relMovesToPoints(*relMoves);
		IPointSet tmp;
		std::pair<IPointSet::const_iterator,bool> setRet;
			// pruefe via einfuegen, ob alle punkte unique sind
		for (IPointVec::const_iterator it = pData.begin(); retVal 
				&& it != pData.end(); it++) {
			setRet = tmp.insert(*it);
			retVal = setRet.second;
		}
		return retVal;
	}
	
		// muss nur pruefen, ob eine struktur da is, da ein relativer
		// move string immer connected ist.
	bool 
	LatticeProtein::isConnected() const {
		bool retVal = relMoves != NULL;	
		/*	IPointVec::const_iterator act = pData->begin(), last = act;
			for (act++; retVal && act != pData->end(); last = act++) {
				retVal = lattice->areNeighbored(*last, *act);
			}*/
		return retVal;
	}
	
	std::string 
	LatticeProtein::getAbsoluteMoveString() const {
		return lattice->getString(getAbsoluteMoves());
	}

	std::string 
	LatticeProtein::getRelativeMoveString() const {
		return lattice->getString(getRelativeMoves());
	}	

	MoveSequence 
	LatticeProtein::getAbsoluteMoves() const {
		assertell (relMoves != NULL && lattice != NULL, 
			"no structure or lattice model available");
		return lattice->relMovesToAbsMoves(*relMoves);
	}

	MoveSequence 
	LatticeProtein::getRelativeMoves() const {
		assertell (relMoves != NULL, "no structure available");
		return *relMoves;
	}	

	// abstract functions
	DPointVec	
	LatticeProtein::get3Ddata() const  {
		assertell (relMoves != NULL, "no structure available");
		IPointVec iVec= (lattice->relMovesToPoints(*relMoves));
		DPointVec dVec;
		for (IPointVec::const_iterator it = iVec.begin();it != iVec.end(); it++)
			dVec.push_back((DblPoint)*it);
		return dVec;
	}
	
	double 
	LatticeProtein::getDRMSD(const BackboneStructure3D& other) const {
		DPointVec this3Ddata = get3Ddata(), other3Ddata = other.get3Ddata();

		assertell(this3Ddata.size() == other3Ddata.size(),
		 "Both lattice proteins have to have the same length.");

		double rmsd = 0.0;

		for(size_t i=0; i < this3Ddata.size()-1; i++) {
			for(size_t j=i+1; j < this3Ddata.size(); j++) {
				rmsd += std::pow( this3Ddata[i].distance(this3Ddata[j])
						 - other3Ddata[i].distance(other3Ddata[j]), 2 );
			}
		}

        rmsd /= (this3Ddata.size() * (this3Ddata.size()-1) /2);

        return std::sqrt(rmsd);

	}

	Sequence 
	LatticeProtein::getSequence() const {
		return *sequence;
	}

	Structure 
	LatticeProtein::getStructure() const {
		return (Structure)getRelativeMoves();
	}

	double	
	LatticeProtein::getEnergy() const {
		assertell(lattice != NULL, "no lattice model available");
		double retVal = 0.0;
		IPointVec pData = lattice->relMovesToPoints(*relMoves);
		IPointVec::size_type i=0,j=0;
		for (i=0; i < pData.size(); i++) {
			for (j = i+2; j < pData.size(); j++) {
				if (lattice->areNeighbored(pData.at(i), pData.at(j))) {
					retVal += energy->getContactEnergy(	sequence->at(i),
														sequence->at(j));
				}
			}
		}
		return retVal;
	}

	bool 
	LatticeProtein::isValid() const {
		return isConnected() && isSelfAvoiding();
	}
	
	std::string	
	LatticeProtein::getStringRepresentation() const {
		assertell (energy != NULL, "no energy function available");
		return	getAbsoluteMoveString()
				+"("+(energy->getAlphabet()->getString((*sequence)))+")";
	}

}	// namespace ell
