#include <ell/LatticeProteinPivotMoveState.hh>
#include <cmath>	// for getFitness()
#include <ell/CLatticeProteinPivotMoveState.hh>
#include <ell/assertell.hh>

namespace ell
{

	const size_t LatticeProteinPivotMoveState::INVALID_INDEX = INT_MAX;
	
	const size_t 
		LatticeProteinPivotMoveState::RandomNeighborList::ItState::SWITCH_MODE_THRESHOLD = 100;
//////////////////////////////////////////////////////////
// LatticeProteinPivotMoveState
//////////////////////////////////////////////////////////

	LatticeProteinPivotMoveState::LatticeProteinPivotMoveState(
	 const LatticeModel* lattice, const ContactEnergyFunction* energy,
	 const std::string& seq, const std::string& moveString,
	 const bool isAbsoluteMove)
	 : latProt(LatticeProtein(lattice, energy, seq, moveString,
							   isAbsoluteMove)),
	   pivotMoveSet(lattice)
	{
		assertell(latProt.isValid(),
		 "LatticeProteinPivotMoveState has to be a valid state.");
	}

	LatticeProteinPivotMoveState::LatticeProteinPivotMoveState(
				const LatticeProteinPivotMoveState& latProtPMState2)
	 : latProt(latProtPMState2.latProt),
	   pivotMoveSet(latProtPMState2.latProt.getLatticeModel()) {
	}

	LatticeProteinPivotMoveState::~LatticeProteinPivotMoveState() {
	}

	LatticeProteinPivotMoveState::PivotMove
	LatticeProteinPivotMoveState::firstPivotMove() const {
			// first move has to be applied at position 1
		return nextPivotMove(PivotMove(0, pivotMoveSet.getMoveNumber()));
	}
		
	LatticeProteinPivotMoveState::PivotMove
	LatticeProteinPivotMoveState::nextPivotMove(const PivotMove lastPivotMove) const {
		if (lastPivotMove.second+1 < pivotMoveSet.getMoveNumber()) {
			return PivotMove(lastPivotMove.first, lastPivotMove.second+1);
		}
			// if the move index reached the number of available moves,
			// increase the move position by one
		else {
			if (lastPivotMove.first+1 < latProt.getStructure().size()) {
				return PivotMove(lastPivotMove.first+1, 0);
			}
			// no more pivot moves possible
			else {
				return PivotMove(0,INVALID_INDEX);
			}
		}
	}

	bool	
	LatticeProteinPivotMoveState::operator== (const State& state2) const {
		return latProt == dynamic_cast<
			const LatticeProteinPivotMoveState&>(state2).latProt;
	}

	bool
	LatticeProteinPivotMoveState::operator!= (const State& state2) const {
		return latProt != dynamic_cast<
			const LatticeProteinPivotMoveState&>(state2).latProt;
	}

	int	
	LatticeProteinPivotMoveState::getFitness() const {
			// rounds fitness and casts it an int-value
		double fitness = latProt.getEnergy();
		if (fitness >= 0)
			return static_cast<int>(floor(fitness + 0.5));
		else
			return static_cast<int>(ceil(fitness - 0.5));
	}

	unsigned int	
	LatticeProteinPivotMoveState::getMinimalDistance(
		 const State& _state2) const {
		LatticeProteinPivotMoveState state2 = 
		 dynamic_cast<const LatticeProteinPivotMoveState&>(_state2);

		Structure state1Struct = latProt.getStructure();
		Structure state2Struct = state2.latProt.getStructure();

		assertell(state2Struct.size() == state1Struct.size(),
		 "Both lattice proteins have to have the same length.");
			
		unsigned int distance = 0;
		for (size_t i = 0; i < state1Struct.size(); i++){
			if (state1Struct[i] != state2Struct[i]) {
				distance += 1;
			}
		}
		return distance;
	}

	LatticeProteinPivotMoveState* 
	LatticeProteinPivotMoveState::clone() const {
		return new LatticeProteinPivotMoveState(*this);
	}
	
	std::string 
	LatticeProteinPivotMoveState::getStateStringRepresentation()
		     const{
		return latProt.getStringRepresentation();
	}

	CState*
	LatticeProteinPivotMoveState::compressedClone() const {
		return new CLatticeProteinPivotMoveState(
		 latProt.getLatticeModel(), latProt.getContactEnergyFunction(),
		 latProt.getSequence(), latProt.getStructure());
	}

	void
	LatticeProteinPivotMoveState::compress(CState& cState) const {
		dynamic_cast<CLatticeProteinPivotMoveState&>(cState)
		 = CLatticeProteinPivotMoveState(
			latProt.getLatticeModel(), latProt.getContactEnergyFunction(),
			latProt.getSequence(), latProt.getStructure());
	}

	State::NeighborListPtr 
	LatticeProteinPivotMoveState::getNeighborList() const {
		return NeighborListPtr(new NeighborList(this));
	}

	State::NeighborListPtr
	LatticeProteinPivotMoveState::getRandomNeighborList()
				const {
		return NeighborListPtr(new RandomNeighborList(this));
	}

	// NeighborList
	LatticeProteinPivotMoveState::NeighborList::NeighborList(
	 			const LatticeProteinPivotMoveState* _origin) 
	 : origin(_origin) 
	{
	}

	LatticeProteinPivotMoveState::NeighborList::~NeighborList() {
	}

	State* 
	LatticeProteinPivotMoveState::NeighborList::first(
		      State::NeighborList::ItState** itstate) const {
		if (*itstate != NULL) delete *itstate;

		ItState* myitstate = new ItState();
		*itstate = myitstate;

		LatticeProteinPivotMoveState* latProtState = NULL;

		PivotMove firstMove = origin->firstPivotMove();

		if (firstMove != PivotMove(0,INVALID_INDEX)) {
			latProtState = new LatticeProteinPivotMoveState(*origin);
			latProtState->pivotMoveSet.applyMoveInPlace(&latProtState->latProt,
			 firstMove.first, firstMove.second);
			myitstate->lastMove = firstMove;
				// if applied pivot move generates an invalid neighbor,
				// call next(..) to iterate neighbors until a VALID one is
				// found which will be returned
			if (!(latProtState->latProt.isValid())) {
				return next(*itstate, latProtState);
			}
		}

		return latProtState;
	}

	State* 
	LatticeProteinPivotMoveState::NeighborList::next(
		      State::NeighborList::ItState* itstate, State* elem) const {
		assertell(elem != NULL, "Elem is not allowed to be NULL");

		ItState* myitstate = static_cast<ItState*>(itstate);

			// find next move
		PivotMove nextMove = origin->nextPivotMove(myitstate->lastMove);
			// no more moves possible
		if (nextMove == PivotMove(0,INVALID_INDEX)) {
			return NULL;
		}
		else {
			LatticeProteinPivotMoveState* latProtState =
			 dynamic_cast<LatticeProteinPivotMoveState*>(elem);
				// set latProtState to the state of origin of the neighbors
			latProtState->latProt = origin->latProt;
				// perform next move
			latProtState->pivotMoveSet.applyMoveInPlace(&latProtState->latProt,
			 nextMove.first, nextMove.second);
			myitstate->lastMove = nextMove;
				// if applied pivot move generates an invalid neighbor,
				// call next(..) to iterate neighbors until a VALID one is
				// found which will be returned
			if (!(latProtState->latProt.isValid())) {
				return next(itstate, latProtState);
			}
			return latProtState;
		}
	}

	// RandomNeighborList
	LatticeProteinPivotMoveState::RandomNeighborList::RandomNeighborList(
	 		const LatticeProteinPivotMoveState* _origin)
	 : origin(_origin) 
	{
	}

	LatticeProteinPivotMoveState::RandomNeighborList::~RandomNeighborList() {
	}

	void 
	LatticeProteinPivotMoveState::RandomNeighborList::switchMode(
			ItState* itstate) const {
		ItState* myitstate = static_cast<ItState*>(itstate);

		PivotMove nextMove = origin->firstPivotMove();

		// loop while moves exist
		while (nextMove != PivotMove(0,INVALID_INDEX)) {
			// add all unchosen moves to unchosenMoves
			if (myitstate->chosenMoves.find(nextMove) == 
			     myitstate->chosenMoves.end()) {
				myitstate->unchosenMoves.push_back(nextMove);
			}
			nextMove = origin->nextPivotMove(nextMove);
		}
	}

	State* 
	LatticeProteinPivotMoveState::RandomNeighborList::first(
		      State::NeighborList::ItState** itstate) const {
		if (*itstate != NULL) delete *itstate;

		*itstate = new ItState();
		LatticeProteinPivotMoveState* latProtState = 
		 new LatticeProteinPivotMoveState(*origin);

		return next(*itstate, latProtState);
	}

	State* 
	LatticeProteinPivotMoveState::RandomNeighborList::next(
		      State::NeighborList::ItState* itstate, State* elem) const {
		assertell(elem != NULL, "Elem is not allowed to be NULL");

		ItState* myitstate = static_cast<ItState*>(itstate);
		LatticeProteinPivotMoveState* latProtState =
		 dynamic_cast<LatticeProteinPivotMoveState*>(elem);

		PivotMove nextMove;

		if (myitstate->switchModeValue > 0) {
				// position and index of the next pivot move
			size_t pos, moveIndex;
			bool foundValidMove = false;

			while (!foundValidMove) {
					// move position has to be in [1, structure length)
				pos = (random() % (origin->latProt.getStructure().size()-1))+1;
					// moveIndex has to be in [0, number of available moves)
				moveIndex = random() % origin->pivotMoveSet.getMoveNumber();

				nextMove = PivotMove(pos, moveIndex);
					// move has already been chosen
				if (!myitstate->chosenMoves.insert(nextMove).second) {
					myitstate->switchModeValue--;
					if (myitstate->switchModeValue == 0) {
						switchMode(myitstate);
						break;
					}
				}
					// move hasn't been chosen yet
				else {
						// set latProtState to the state of origin of the
						// neighbors
					latProtState->latProt = origin->latProt;
					latProtState->pivotMoveSet.applyMoveInPlace(&latProtState->latProt,
					 nextMove.first, nextMove.second);
						// if applied pivot move generates an invalid neighbor,
						// iterate neighbors until a VALID one is found which
						// will be returned
					foundValidMove = latProtState->latProt.isValid();
				}	
			}
		}
		if (myitstate->switchModeValue == 0) {
			size_t n = myitstate->unchosenMoves.size();
			if (n == 0) {
				return NULL;
			}
			else {
				size_t i = random() % n;
					// set latProtState to the state of origin of the neighbors
				latProtState->latProt = origin->latProt;
				nextMove = myitstate->unchosenMoves[i];
				latProtState->pivotMoveSet.applyMoveInPlace(&latProtState->latProt,
				 nextMove.first, nextMove.second);
				myitstate->unchosenMoves[i] =
				 myitstate->unchosenMoves[n-1];
				myitstate->unchosenMoves.resize(n-1);
					// if applied pivot move generates an invalid neighbor,
					// call next(..) to iterate neighbors until a VALID one is
					// found which will be returned
				if (!(latProtState->latProt.isValid())) {
					return next(itstate, latProtState);
				}
			}
		}
		return latProtState;
	}

} // namespace ell
