#include <ell/RNAFreeEnergySingleMoveState.hh>
#include <cmath>	// for getFitness()
#include <ell/CRNAFreeEnergySingleMoveState.hh>
#include <ell/assertell.hh>

namespace ell
{

	const size_t 
		RNAFreeEnergySingleMoveState::RandomNeighborList::ItState::SWITCH_MODE_THRESHOLD = 100;
//////////////////////////////////////////////////////////
// RNAFreeEnergySingleMoveState
//////////////////////////////////////////////////////////

	RNAFreeEnergySingleMoveState::RNAFreeEnergySingleMoveState(
	 			const std::string& rnaSeqStr, 
	 			const std::string& rnaStructBracketNotStr)
	 :	rnaFreeEnergy(RNAFreeEnergy(rnaSeqStr, rnaStructBracketNotStr)) 
	{
		assertell(rnaFreeEnergy.isValid(),
		 "RNAFreeEnergySingleMoveState has to be a valid state.");
	}

	RNAFreeEnergySingleMoveState::RNAFreeEnergySingleMoveState(
				const RNAFreeEnergySingleMoveState& rnaFESMS)
	 : rnaFreeEnergy(rnaFESMS.rnaFreeEnergy) {
	}

	RNAFreeEnergySingleMoveState::~RNAFreeEnergySingleMoveState() {
	}

	RNAFreeEnergySingleMoveState::SingleMove 
	RNAFreeEnergySingleMoveState::firstValidSingleMove() const {
		return searchForNextValidSingleMove(0,1);
	}

	RNAFreeEnergySingleMoveState::SingleMove 
	RNAFreeEnergySingleMoveState:: nextValidSingleMove( 
			const SingleMove lastSingleMove) const 
	{
		// if lastSingleMove.first is an opening bond, this bond was already
		// modified in the last move
		// closing bonds at lastSingleMove.first are not for interest
		if (rnaFreeEnergy.getStructure().at(lastSingleMove.first) !=
			 RNAFreeEnergy::getStructureAlphabet()->getElement(".") ) {
			return searchForNextValidSingleMove(lastSingleMove.first+1,
				lastSingleMove.first+2);
		}
		else {
			return searchForNextValidSingleMove(lastSingleMove.first,
				lastSingleMove.second+1);
		}
	}

	RNAFreeEnergySingleMoveState::SingleMove 
	RNAFreeEnergySingleMoveState:: searchForNextValidSingleMove(	size_t pos1, 
								size_t pos2) const 
	{
		assertell(pos1 < pos2,
		 "RNA single move search pos 1 has to be less than pos 2");
		
		// distance between pos1 and pos2 has to be at least
		//  MIN_LOOP_LENGTH
		while ( (pos1+rnaFreeEnergy.getMinLoopLength())
				 < rnaFreeEnergy.getStructure().size()) 
		{
			if ( rnaFreeEnergy.getStructure().at(pos1) ==
				  RNAFreeEnergy::getStructureAlphabet()->getElement(".") )
			{
				for(pos2 = rnaFreeEnergy.nextCompatibleSingleMovePos(pos2-1);
					pos2 < rnaFreeEnergy.getStructure().size();
					pos2 = rnaFreeEnergy.nextCompatibleSingleMovePos(pos2))
				{
					if ((pos2 - pos1) > rnaFreeEnergy.getMinLoopLength() &&
						 rnaFreeEnergy.isAllowedBasePair(pos1, pos2)) {
							// found bases which are able to pair
							// -> exit loop
						return SingleMove(pos1,pos2);
						break;
					}
				}
			}
			else if ( rnaFreeEnergy.getStructure().at(pos1) ==
						RNAFreeEnergy::getStructureAlphabet()->getElement("(") )
			{
				pos2 = rnaFreeEnergy.getClosingBond(pos1);
					// existing bond found -> exit loop
				return SingleMove(pos1,pos2);
			}
				// loop wasn't exited since no next valid single move was found
			pos1++;
			pos2 = pos1;
		}
		// no next valid single move found
		return SingleMove(0,RNAStructure::INVALID_INDEX);
	}

	bool	
	RNAFreeEnergySingleMoveState::operator== (const State& state2) const {
		return rnaFreeEnergy == (dynamic_cast<
			const RNAFreeEnergySingleMoveState&>(state2)).rnaFreeEnergy;
	}

	bool
	RNAFreeEnergySingleMoveState::operator!= (const State& state2) const {
		return rnaFreeEnergy != (dynamic_cast<
			const RNAFreeEnergySingleMoveState&>(state2)).rnaFreeEnergy;
	}

	int	
	RNAFreeEnergySingleMoveState::getFitness() const {
			// rounds fitness and multiplies it with 100
		double fitness = 100 * rnaFreeEnergy.getEnergy();
		if (fitness >= 0)
			return static_cast<int>(floor(fitness + 0.5));
		else
			return static_cast<int>(ceil(fitness - 0.5));
	}

	unsigned int	
	RNAFreeEnergySingleMoveState::getMinimalDistance(
		 const State& _state2) const {
		RNAFreeEnergySingleMoveState state2 = 
		 dynamic_cast<const RNAFreeEnergySingleMoveState&>(_state2);

		Structure state1Struct = rnaFreeEnergy.getStructure();
		Structure state2Struct = state2.rnaFreeEnergy.getStructure();

		assertell(state2Struct.size() == state1Struct.size(),
		"Length of both RNA has to be equal to compute the distance.");
			
		unsigned int distance = 0;
		for (size_t i = 0; i < state1Struct.size(); i++){
			// bond maintained or bond opened and replaced by another one
			if (state1Struct[i] ==
				 RNAFreeEnergy::getStructureAlphabet()->getElement("(") &&
			   state2Struct[i] ==
				 RNAFreeEnergy::getStructureAlphabet()->getElement("(")) {
				// bond opened and replaced by another one
				if (rnaFreeEnergy.getClosingBond(i) !=
				     state2.rnaFreeEnergy.getClosingBond(i)) {
					distance += 2;
				}
			}
			// bond opened or closed
			else if (state1Struct[i] ==
					   RNAFreeEnergy::getStructureAlphabet()->getElement("(") ||
					  state2Struct[i] ==
					   RNAFreeEnergy::getStructureAlphabet()->getElement("(")) {
				distance += 1;
			}
		}
		return distance;
	}

	RNAFreeEnergySingleMoveState* 
	RNAFreeEnergySingleMoveState::clone() const {
		return new RNAFreeEnergySingleMoveState(*this);
	}
	
	std::string 
	RNAFreeEnergySingleMoveState::getStateStringRepresentation()
		     const{
		return rnaFreeEnergy.getStringRepresentation();
	}

	CState*
	RNAFreeEnergySingleMoveState::compressedClone() const {
		return new CRNAFreeEnergySingleMoveState(
		 rnaFreeEnergy.getSequence(),rnaFreeEnergy.getStructure());
	}

	void
	RNAFreeEnergySingleMoveState::compress(CState& cState) const {
		dynamic_cast<CRNAFreeEnergySingleMoveState&>(cState)
		 = CRNAFreeEnergySingleMoveState(
			rnaFreeEnergy.getSequence(),rnaFreeEnergy.getStructure());
	}

	State::NeighborListPtr 
	RNAFreeEnergySingleMoveState::getNeighborList() const {
		return NeighborListPtr(new NeighborList(this));
	}

	State::NeighborListPtr
	RNAFreeEnergySingleMoveState::getRandomNeighborList()
				const {
		return NeighborListPtr(new RandomNeighborList(this));
	}

	// NeighborList
	RNAFreeEnergySingleMoveState::NeighborList::NeighborList(
	 			const RNAFreeEnergySingleMoveState* _origin) 
	 : origin(_origin) 
	{
	}

	RNAFreeEnergySingleMoveState::NeighborList::~NeighborList() {
	}

	State* 
	RNAFreeEnergySingleMoveState::NeighborList::first(
		      State::NeighborList::ItState** itstate) const {
		if (*itstate != NULL) delete *itstate;

		ItState* myitstate = new ItState();
		*itstate = myitstate;
		
		RNAFreeEnergySingleMoveState* rnaState = NULL;

		SingleMove firstMove = origin->firstValidSingleMove();

		if (firstMove != SingleMove(0,RNAStructure::INVALID_INDEX)) {
			rnaState = new RNAFreeEnergySingleMoveState(*origin);
			rnaState->rnaFreeEnergy.applySingleMoveInPlace(firstMove);
			myitstate->lastMove = firstMove;
		}
		return rnaState;
	}

	State* 
	RNAFreeEnergySingleMoveState::NeighborList::next(
		      State::NeighborList::ItState* itstate, State* elem) const {
		assertell(elem != NULL, "Elem is not allowed to be NULL");

		ItState* myitstate = static_cast<ItState*>(itstate);

		// find next move
		SingleMove nextMove = origin->nextValidSingleMove(myitstate->lastMove);
		// no more moves possible
		if (nextMove == SingleMove(0,RNAStructure::INVALID_INDEX)) {
			return NULL;
		}
		else {
			RNAFreeEnergySingleMoveState* rnaState =
			 dynamic_cast<RNAFreeEnergySingleMoveState*>(elem);
			// undo last move
			rnaState->rnaFreeEnergy.applySingleMoveInPlace(myitstate->lastMove);
			// perform next move
			rnaState->rnaFreeEnergy.applySingleMoveInPlace(nextMove);
			myitstate->lastMove = nextMove;
			return rnaState;
		}
	}

	// RandomNeighborList
	RNAFreeEnergySingleMoveState::RandomNeighborList::RandomNeighborList(
	 		const RNAFreeEnergySingleMoveState* _origin)
	 : origin(_origin) 
	{
	}

	RNAFreeEnergySingleMoveState::RandomNeighborList::~RandomNeighborList() {
	}

	RNAFreeEnergySingleMoveState::SingleMove 
	RNAFreeEnergySingleMoveState:: RandomNeighborList::getOrderedSingleMove(
			size_t pos1,
	 		size_t pos2) const 
	{
		return (pos1 < pos2) ?
			SingleMove(pos1, pos2) : SingleMove(pos2, pos1);
	}

	void 
	RNAFreeEnergySingleMoveState::RandomNeighborList::switchMode(
			ItState* itstate) const 
	{
		ItState* myitstate = static_cast<ItState*>(itstate);

		SingleMove nextMove = origin->firstValidSingleMove();

		// loop while valid moves exist
		while (nextMove != SingleMove(0,RNAStructure::INVALID_INDEX)) {
			// add all valid and unchosen moves to unchosenValidMoves
			if (myitstate->chosenMoves.find(nextMove) == 
			     myitstate->chosenMoves.end()) {
				myitstate->unchosenValidMoves.push_back(nextMove);
			}
			nextMove = origin->nextValidSingleMove(nextMove);
		}
	}

	State* 
	RNAFreeEnergySingleMoveState::RandomNeighborList::first(
		      State::NeighborList::ItState** itstate) const 
	{
		if (*itstate != NULL) delete *itstate;

		*itstate = new ItState();
		RNAFreeEnergySingleMoveState* rnaState = 
		 new RNAFreeEnergySingleMoveState(*origin);

		return next(*itstate, rnaState);
	}

	State* 
	RNAFreeEnergySingleMoveState::RandomNeighborList::next(
		      State::NeighborList::ItState* itstate, State* elem) const 
	{
		assertell(elem != NULL, "Elem is not allowed to be NULL");

		ItState* myitstate = static_cast<ItState*>(itstate);
		RNAFreeEnergySingleMoveState* rnaState =
		 dynamic_cast<RNAFreeEnergySingleMoveState*>(elem);

		if (myitstate->switchModeValue > 0) {
			// positions the move will be applied to
			size_t pos1, pos2;
			SingleMove nextMove;
			bool foundValidMove = false;

			while (!foundValidMove) {
				pos1 = random() % origin->rnaFreeEnergy.getStructure().size();
				pos2 = random() % origin->rnaFreeEnergy.getStructure().size();
				if (pos1 == pos2) {
					continue;
				}
				nextMove = getOrderedSingleMove(pos1, pos2);
				// move has already been chosen
				if (!myitstate->chosenMoves.insert(nextMove).second) {
					myitstate->switchModeValue--;
					if (myitstate->switchModeValue == 0) {
						switchMode(myitstate);
						break;
					}
				}
				// move hasn't been chosen yet and is valid
				else if (origin->rnaFreeEnergy.
					 isValidSingleMove(nextMove)) {
						// set rnaState to the state of origin of the neighbors
					rnaState->rnaFreeEnergy = origin->rnaFreeEnergy;
					rnaState->rnaFreeEnergy.applySingleMoveInPlace(
					 nextMove);
					foundValidMove = true;
				}	
			}
		}
		if (myitstate->switchModeValue == 0) {
			size_t n = myitstate->unchosenValidMoves.size();
			if (n == 0) {
				return NULL;
			}
			else {
				size_t i = random() % n;
					// set rnaState to the state of origin of the neighbors
				rnaState->rnaFreeEnergy = origin->rnaFreeEnergy;
				rnaState->rnaFreeEnergy.applySingleMoveInPlace(
				 myitstate->unchosenValidMoves[i]);
				myitstate->unchosenValidMoves[i] =
				 myitstate->unchosenValidMoves[n-1];
				myitstate->unchosenValidMoves.resize(n-1);
			}
		}
		return rnaState;
	}

} // namespace ell
